// Copyright (C) 2012-2015 ChaosForge Ltd
// http://chaosforge.org/
//
// This file is part of Nova libraries. 
// For conditions of distribution and use, see copying.txt file in root folder.
/**
 * @file time.hh
 * @author Kornel Kisielewicz epyon@chaosforge.org
 * @brief Time related functions
 */

#ifndef NV_CORE_TIME_HH
#define NV_CORE_TIME_HH

#include <nv/common.hh>

namespace nv
{
	/**
	 * Returns the amount of ticks of the processors high resolution 
	 * timer. Currently only supported on GCC and MSVC. Probably only on
	 * some architectures.
	 *
	 * @returns amount of ticks
	 */
	uint64 get_ticks();

	/**
	 * Performs an operating system sleep call.
	 *
	 * @param ms time in milliseconds to sleep
	 */
	void sleep( uint32 ms );

	/**
	 * Get millisecond count based on std::clock
	 */
	uint32 get_cpu_ms();

	/**
	 * Get microsecond count based on std::clock
	 */
	uint64 get_cpu_us();

	/**
	 * Get millisecond count based on system counter
	 */
	uint32 get_system_ms();

	/**
	 * Get microsecond count based on system counter
	 */
	uint64 get_system_us();

	struct cpu_ms_timer 
	{ 
		typedef uint32 value_type;
		static const value_type second = 1000;    
		value_type operator()() { return get_cpu_ms(); } 
	};

	struct cpu_us_timer 
	{ 
		typedef uint64 value_type;
		static const value_type second = 1000000; 
		uint64 operator()() { return get_cpu_us(); } 
	};

	struct system_ms_timer 
	{ 
		typedef uint32 value_type;
		static const value_type second = 1000;    
		value_type operator()() { return get_system_ms(); } 
	};

	struct system_us_timer 
	{ 
		typedef uint64 value_type;
		static const value_type second = 1000000; 
		value_type operator()() { return get_system_us(); } 
	};
	
	/**
	 * Timer template class
	 */
	template < class Timer >
	class timer_class
	{
	public:
		typedef typename Timer::value_type value_type;

		timer_class()	: last( Timer()() ) {}
		void mark()	
		{
			value_type now = Timer()();
			stored = now - last;
			last = now;
		}
		value_type elapsed() const
		{
			return stored;
		}
	private:
		value_type last;
		value_type stored;
	};

	/**
	 * FPS template class
	 */
	template < class Timer >
	class fps_counter_class
	{
	public:
		typedef typename Timer::value_type value_type;

		fps_counter_class() : m_frames(1), m_last(0), m_value(0.0f) {}
		bool tick()
		{
			value_type now = Timer()();
			if ( now - m_last >= Timer::second )
			{
				m_value = (static_cast<float>(m_frames) /
					static_cast<float>(now - m_last))*Timer::second;
				m_frames = 1;
				m_last = now;
				return true;
			}
			m_frames++;
			return false;
		}
		f32 fps() const
		{
			return m_value;
		}
	private:
		uint32     m_frames;
		value_type m_last;
		f32        m_value;
	};

	typedef timer_class< system_ms_timer > ms_timer;
	typedef timer_class< system_us_timer > us_timer;
	typedef fps_counter_class< system_ms_timer > fps_counter;

} // namespace nv

#endif // NV_CORE_TIME_HH
